/*======================================================================================

	Program:		Range Imaging Toolkit (RITK)
	Module:			ritk3DCuda3DPitchedMemoryImportImageContainer.txx
	Version:		$Revision: 2883 $
	Date:			$Date: 2012-04-02 17:01:22 +0200 (Mo, 02 Apr 2012) $

	Copyright (c) 2010-2011
	Pattern Recognition Lab, 
    University Erlangen-Nuremberg, Germany
    All rights reserved.

	This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See RITKCopyright.txt for more information.

	http://www5.informatik.uni-erlangen.de/research/software/ritk

========================================================================================*/

#include <cuda_runtime.h>

#include "ritkCuda3DPitchedMemoryImportImageContainer.h"

#include "ritkDebugManager.h"

namespace ritk
{

//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>
::Cuda3DPitchedMemoryImportImageContainer()
{ 
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>
::~Cuda3DPitchedMemoryImportImageContainer()
{
	this->DeallocateManagedMemory();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
PrintSelf(std::ostream& os, itk::Indent indent) const
{
	os << indent << "Ptr @" << m_CudaMemory.front().first.ptr << std::endl;
	os << indent << "Pitch=" << m_CudaMemory.front().first.pitch << std::endl;
	os << indent << "XSize=" << m_CudaMemory.front().first.xsize << std::endl;
	os << indent << "YSize=" << m_CudaMemory.front().first.ysize << std::endl;
	Superclass::PrintSelf(os,indent);
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
const cudaExtent &
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
GetExtent() const
{
	return m_CudaMemory.front().second;
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
cudaExtent &
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
GetExtent()
{
	return m_CudaMemory.front().second;
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
typename Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::Element*
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
AllocateElements(ElementIdentifier size) const
{
	// Sanity checks. This ensures that we can operate on the internal 2-D size parameters
	if ( Superclass::m_ContainerSize[0] == 0 || Superclass::m_ContainerSize[1] == 0 || Superclass::m_ContainerSize[2] == 0 )
	{
		LOG_DEB("Invalid width/height/depth detected: did you forget to set them via SetContainerSize()?");
		itkExceptionMacro(<<"Invalid width/height/depth detected: did you forget to set them via SetContainerSize()?");
	}
	if ( Superclass::m_ContainerSize[0]*Superclass::m_ContainerSize[1]*Superclass::m_ContainerSize[2] != size )
	{
		LOG_DEB("Mismatch between widht/height/depth and requested size detected: " << Superclass::m_ContainerSize[0] << "*" << Superclass::m_ContainerSize[1] << "*" << Superclass::m_ContainerSize[2] << "=" << (Superclass::m_ContainerSize[0]*Superclass::m_ContainerSize[1]*Superclass::m_ContainerSize[2]) << " vs. " << size);
		itkExceptionMacro(<<"Mismatch between widht/height/depth and requested size detected: " << Superclass::m_ContainerSize[0] << "*" << Superclass::m_ContainerSize[1] << "*" << Superclass::m_ContainerSize[2] << "=" << (Superclass::m_ContainerSize[0]*Superclass::m_ContainerSize[1]*Superclass::m_ContainerSize[2]) << " vs. " << size);
	}

	// Execute in the global CUDA context. Note that we can't use the size argument directly. Instead, we use the internal 3-D size!
	CudaBlockedConstMethodCallMacro(AllocateCudaMemory);

	// Call the superclass' implementation. This also returns the correct host memory pointer
	return Superclass::AllocateElements(size);
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
DeallocateManagedMemory()
{
	// Execute in the global CUDA context
	CudaBlockedMethodCallMacro(DeallocateCudaMemory);

	// Call the superclass' implementation to deallocate host memory
	Superclass::DeallocateManagedMemory();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
SynchronizeDevice()
{
	// Execute in the global CUDA context
	CudaBlockedMethodCallMacro(CopyHostToDevice);

	// Synchronize MTimes
	Superclass::SynchronizeDevice();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
SynchronizeHost()
{
	// Execute in the global CUDA context
	CudaBlockedMethodCallMacro(CopyDeviceToHost);

	// Synchronize MTimes
	Superclass::SynchronizeHost();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
DeepCopy(const ImportImageContainerType *Container)
{
	// Cast and pass
	ImportImageContainerType *NonConstContainer = const_cast<ImportImageContainerType*>(Container);
	if ( !NonConstContainer )
	{
		LOG_DEB("Could not remove const qualifier!");
		itkExceptionMacro(<<"Could not remove const qualifier!");
	}
	DeepCopy(NonConstContainer);
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
DeepCopy(ImportImageContainerType *Container)
{
	// Call the superclass implementation first. This will ensure correct memory allocation and host memory copy.
	Superclass::DeepCopy(Container);

	// Check if we have a cuda container at all. If no, the GenericImportImageContainer should call a Synchronize() triggered by Superclass::DeepCopy() on us
	Superclass *CudaContainer = dynamic_cast<Superclass*>(Container);
	if ( !CudaContainer )
		return;

	// Check for MTimes: we do not need to copy device memory that is out-of-date. Eventually, this has to be resolved by a synch!
	if ( CudaContainer->GetDeviceMTime() < CudaContainer->GetHostMTime() )
	{
		GenericImportImageContainerType::m_DeviceMTime = CudaContainer->GetDeviceTimeStamp();
		return;
	}

	// Device copy
	m_TmpCudaContainer = CudaContainer;
	CudaBlockedMethodCallMacro(DeepCopyCudaMemory);
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
CopyToMappedResource(void *d_ptr, size_t size) const
{
	itkExceptionMacro(<<"Copying to mapped resources not supported by an array container!");
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
cudaPitchedPtr&
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
GetCudaPitchedMemoryPointer()
{
	GenericImportImageContainerType::DeviceModified();
	return m_CudaMemory.front().first;
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
const cudaPitchedPtr&
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
GetCudaPitchedMemoryPointer() const
{
	return m_CudaMemory.front().first;
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
AllocateCudaMemory() const
{
	// Allocate. Note that a correct size is checked in AllocateElements()
	cudaExtent VolumeSize = make_cudaExtent(Superclass::m_ContainerSize[0]*sizeof(Element),Superclass::m_ContainerSize[1],Superclass::m_ContainerSize[2]);
	cudaPitchedPtr Ptr;
	ritkCudaSafeCall(cudaMalloc3D((cudaPitchedPtr*)&Ptr, VolumeSize));

	// Save the allocated memory at the end of our internal list such that it won't be deleted through the itk (de)allocate scheme
	std::pair<cudaPitchedPtr, cudaExtent> tmp;
	tmp.first = Ptr;
	tmp.second = VolumeSize;
	m_CudaMemory.push_back(tmp);
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
DeallocateCudaMemory()
{
	// Sanity check
	if ( m_CudaMemory.size() == 0 )
	{
		return;
	}

	// Delete the first element in our internal list
	cudaPitchedPtr Ptr = m_CudaMemory.front().first;
	ritkCudaSafeCall(cudaFree(Ptr.ptr));
	m_CudaMemory.pop_front();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
CopyDeviceToHost()
{
	// Copy data from device to host
	cudaExtent Extent = make_cudaExtent(Superclass::m_ContainerSize[0]*sizeof(Element),Superclass::m_ContainerSize[1],Superclass::m_ContainerSize[2]);
    cudaMemcpy3DParms CopyParams = {0};
	CopyParams.dstPtr = make_cudaPitchedPtr(GenericImportImageContainerType::GetBufferPointer(), Superclass::m_ContainerSize[0]*sizeof(Element), Superclass::m_ContainerSize[0], Superclass::m_ContainerSize[1]);
    CopyParams.srcPtr = this->GetCudaPitchedMemoryPointer();
    CopyParams.extent = Extent;
    CopyParams.kind  = cudaMemcpyDeviceToHost;
    ritkCudaSafeCall(cudaMemcpy3D(&CopyParams));
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
CopyHostToDevice()
{
	// Copy data from host to device
	cudaExtent Extent = make_cudaExtent(Superclass::m_ContainerSize[0]*sizeof(Element),Superclass::m_ContainerSize[1],Superclass::m_ContainerSize[2]);
    cudaMemcpy3DParms CopyParams = {0};
	CopyParams.srcPtr = make_cudaPitchedPtr(GenericImportImageContainerType::GetBufferPointer(), Superclass::m_ContainerSize[0]*sizeof(Element), Superclass::m_ContainerSize[0], Superclass::m_ContainerSize[1]);
    CopyParams.dstPtr = this->GetCudaPitchedMemoryPointer();
    CopyParams.extent = Extent;
    CopyParams.kind  = cudaMemcpyHostToDevice;
    ritkCudaSafeCall(cudaMemcpy3DAsync(&CopyParams));
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement>
void
Cuda3DPitchedMemoryImportImageContainer<TElementIdentifier, TElement>::
DeepCopyCudaMemory()
{
	// 3DPitched -> 3DPitched
	const Self *PitchedContainer = dynamic_cast<const Self*>(m_TmpCudaContainer);
	if ( PitchedContainer )
	{
		cudaExtent VolumeSize = make_cudaExtent(Superclass::m_ContainerSize[0]*sizeof(Element),Superclass::m_ContainerSize[1],Superclass::m_ContainerSize[2]);
		cudaMemcpy3DParms CopyParams = {0};
		CopyParams.srcPtr = PitchedContainer->GetCudaPitchedMemoryPointer();
		CopyParams.dstPtr = this->GetCudaPitchedMemoryPointer();
		CopyParams.extent = VolumeSize;
		CopyParams.kind  = cudaMemcpyDeviceToDevice;
		ritkCudaSafeCall(cudaMemcpy3DAsync(&CopyParams));

		// Update the timestamps. Take care if host and device are already synchronized!
		GenericImportImageContainerType::DeviceModified();
		if ( m_TmpCudaContainer->GetDeviceTimeStamp() == m_TmpCudaContainer->GetHostTimeStamp() )
			GenericImportImageContainerType::m_HostMTime = GenericImportImageContainerType::m_DeviceMTime;

		// Clean up
		m_TmpCudaContainer = NULL;
		return;
	}

	// 3DArray -> 3DPitched
	typedef Cuda3DArrayImportImageContainer<ElementIdentifier, Element> Cuda3DArrayImportImageContainerType;
	const Cuda3DArrayImportImageContainerType *ArrayContainer = dynamic_cast<const Cuda3DArrayImportImageContainerType*>(m_TmpCudaContainer);
	if ( ArrayContainer )
	{
		cudaExtent VolumeSize = make_cudaExtent(Superclass::m_ContainerSize[0],Superclass::m_ContainerSize[1],Superclass::m_ContainerSize[2]);
		cudaMemcpy3DParms CopyParams = {0};
		CopyParams.srcArray = const_cast<cudaArray*>(ArrayContainer->GetCudaMemoryArray()); // ugly but copy cudaMemcpy3DParms doesn't know const...
		CopyParams.dstPtr = this->GetCudaPitchedMemoryPointer();
		CopyParams.extent = VolumeSize;
		CopyParams.kind  = cudaMemcpyDeviceToDevice;
		ritkCudaSafeCall(cudaMemcpy3DAsync(&CopyParams));

		// Update the timestamps. Take care if host and device are already synchronized!
		GenericImportImageContainerType::DeviceModified();
		if ( m_TmpCudaContainer->GetDeviceTimeStamp() == m_TmpCudaContainer->GetHostTimeStamp() )
			GenericImportImageContainerType::m_HostMTime = GenericImportImageContainerType::m_DeviceMTime;

		// Clean up
		m_TmpCudaContainer = NULL;
		return;
	}

	// We can't handle this container
	LOG_DEB("Invalid container detected: " << m_TmpCudaContainer);
	itkExceptionMacro(<<"Invalid container detected: " << m_TmpCudaContainer);
}


} // ritk